/*                                                                                                            
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or
 * http://eu-egee.org/license.html
 */

/*
 * Copyright (c) 2001 EU DataGrid.                                                                             
 * For license conditions see http://www.eu-datagrid.org/license.html                                          
 *
 * Copyright (c) 2001, 2002 by
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 */

/*!
    \page lcmaps_jobrep.mod Job Repository plugin
 
    \section posixsynopsis SYNOPSIS
    \b lcmaps_jobrep.mod -vomsdir \<path\> -certdir \<path\> [server=\<MySQL hostname/dns-name\>] [dsn=\<DSN-name\>]
    [port=\<port nr.\>] [driver=\<path/file\>] [database=\<database-name\>] [username=\<database username\>]
    [password=\<database password\>] [-jr_config \<path/file\>]
 
    \section posixdesc DESCRIPTION
 
The Job Repository module collects the user, job and user-mapping information when a job has been assigned
to a fabric and handled by LCMAPS. This information is stored into a SQL database. It is essential that 
this information is kept secure due to the nature of the data it stores.

 
    \section posixoptions OPTIONS

    \subsection dns1 dsn=\<DSN-name\>
        This will select the Data Source Name (DSN) that has been set in a odbc.ini file. In the odbc.ini file are one or more Database sources described with the needed settings to connect to a database. (Note: also partial connection settings are supported (3rd party support))

    \subsection server1 server=\<dns-name\>
         specifies the host that has the database to connect to.

    \subsection portnr1 port=\<port nr.\>
         specifies the port number on the host to override the default portnumber

    \subsection driver1 driver=\<path/file\>
         specifies the path and filename of the shared object distributed by the MyODBC package. ODBC needs a driver file to know what kind of database it will be connecting to.

    \subsection database1 database=\<database-name\>
         specifies the name of the database that has to be connected to. One machine can host several databases. The default name to fill in is: "JobRepository".

    \subsection username1 username=\<database username\>
         specifies the database username that the LCMAPS module must use to authorize itself with.

    \subsection password1 password=\<database password\>
         specifies the database password that the LCMAPS module must use to authorize itself with.
                                         (Note: Disadviced to use this in the lcmaps.db file because it is publicly accessable)

    \subsection jrconfig1 -jr_config \<path/file\>
         The path and file specification leads to the root-only-readable file that should contain the database password (or more things like the username) to prevent that a simple user could comprimse the information that is in the database. There is a specific format for this file. All these database parameters must be semi-colon delimited. The same parameters as states above are used. This means that in the lcmaps.db all these parameters could be in this file. Also that is not really adviced due to practical reasons (shifting of servers for example) then you'll just have to edit one file. The lcmaps.db file in this case.
This means that you can mix these parameters as you like between the files. All these parameters are concatinated after each other. When parameters are more then ounce defined the last one in the concatinated string will be effectivly used.
                                         (default: /opt/edg/etc/lcmaps/jobrep_config)

    \subsection vomsdir1 -vomsdir \<path\>
         This is the same directory as the LCMAPS VOMS Extract module looks at. This where the VO certificates are stored.
                                         (default: /etc/grid-security/vomsdir/).
                                         Note: this parameter can not be used in the jobrep_config file.

    \subsection certdir1 -certdir \<path\>
         This is the same directory as the LCMAPS VOMS Extract module looks at. This where the user certificates end up (temporarily)
                                         (default: /etc/grid-security/certificates/).
                                         Note: this parameter can not be used in the jobrep_config file.
*/


/*!
    \file   lcmaps_jobrep.c
    \brief  Plugin to get data for the Job Repository Database
    \author Oscar Koeroo for the EU DataGrid.
*/





/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#include <openssl/x509.h>
#include <openssl/x509_vfy.h>
#include <openssl/asn1.h>

#ifdef LCMAPS_GSI_MODE
#   include <gssapi.h>
#endif

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
//    #include "lcmaps_voms_utils.h" 
#include <lcmaps/lcmaps_vo_data.h>

//    #include "voms_apic.h"
//    #include "globus_gss_assist.h"
//#include "gssapi_openssl.h"

#include <jobrepository/jobrep_api.h>

//#undef TESTBIO
//#define TESTBIO
#ifdef TESTBIO
#    include <openssl/pem.h>
#    include <openssl/bio.h>
#endif


/******************************************************************************
                                Definitions
******************************************************************************/

#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1

//#define SPIT_SQL_ONLY     1


// #define VOMS_BUFFER_SIZE  1024

// What do we want to do with this time presentation?
#define GMT               1
#define LOCAL             2
#define LEAVE_TIME_ITS_OK 3



typedef struct TUnixUid_s
{
   int    uid;
   char * uid_name;
   char * voms_attribute_id;
   char * job_id;
   char * user_id;
} TUnixUid;


typedef struct TUnixGid_s
{
   int    gid;
   char * gid_name;
   int    is_primary;
   char * voms_attribute_id;
   char * job_id;
   char * user_id;
} TUnixGid;





/******************************************************************************
                          Module specific prototypes
******************************************************************************/
static int plugin_run_or_verify(int, lcmaps_argument_t *, int);

void testDatabaseODBC (char * db_name, char * db_user, char * db_pwd);
//static void print_vomsdata(struct vomsdata *);


char * strMangleMySQL (char * strMangle);
int parseRawCertificates (int,
                          lcmaps_argument_t *,
                          char *,
                          TSimpleCert **);

int parseStackOfX509 (STACK_OF(X509) *, TSimpleCert **);


//void print_vomsdata(struct vomsdata *);
//static char *gridmapdir_urlencode(char *);
static struct tm * ASN1_TIME_2_time_t(ASN1_TIME *);
char * sttm_2_char (const struct tm *);

int strnclean (char **s, int bufsize);
int strclean  (char **s);


char * createFQAN(char *, char *, char *, char *);
char * tripletLine2FQAN (char *);

static int lcmaps_get_jobrep_config(const char *);

int parseConfigString (char *);


#ifdef TESTBIO
static STACK_OF(X509) *load_chain(char *certfile);
/*
static char *retmsg[] = { "VERR_NONE", "VERR_NOSOCKET", "VERR_NOIDENT", "VERR_COMM", 
                          "VERR_PARAM", "VERR_NOEXT", "VERR_NOINIT",
                          "VERR_TIME", "VERR_IDCHECK", "VERR_EXTRAINFO",
                          "VERR_FORMAT", "VERR_NODATA", "VERR_PARSE",
                          "VERR_DIR", "VERR_SIGN", "VERR_SERVER", 
                          "VERR_MEM", "VERR_VERIFY", "VERR_IDENT",
                          "VERR_TYPE", "VERR_ORDER" };
*/
#endif

/******************************************************************************
                       Define module specific variables
******************************************************************************/

//static char * jobrep_config  = NULL;
//static char * connStr        = NULL;
static char * jr_db_name     = NULL;
static char * jr_db_user     = NULL;
static char * jr_db_pwd      = NULL;
static char * jr_config_file = NULL;

// static char * certdir = NULL;
// static char * vomsdir = NULL;
// static char   voms_buffer[VOMS_BUFFER_SIZE];

static int    test_jobrep = 0;

/*****************************************************************************/




/******************************************************************************
Function:   testDatabaseODBC
Description:
            Tests if the ODBC connection to the database can be setup.

Parameters:


Returns:
    !NULL              : succes
    NULL               : failure
******************************************************************************/

void testDatabaseODBC (char * db_name, char * db_user, char * db_pwd)
{
    int odbc_test_r_value = 0;

    lcmaps_log      (0,"\tJob Repository enters Test Mode\n");

    odbc_test_r_value = ODBC_Test (db_name, db_user, db_pwd);

    lcmaps_log      (0,"\tODBC Test return value: %d\n", odbc_test_r_value);
}




/******************************************************************************
Function:   parseRawCertificates
Description:
         This will parse the certificate chain and will create a structure 
         from which you can more easiliy extract all the data to insert into 
         the database.
         Note: This EXCLUDES VOMS parsing/extraction

Parameters:
         

Returns:
    !NULL              : succes
    NULL               : failure
******************************************************************************/

//int parseCertificates (int argc,  
//                       lcmaps_argument_t * argv,
//                       char * logstr,
//                       TSimpleCert * MySimpleStack)
int parseRawCertificates (int argc,  
                       lcmaps_argument_t * argv,
                       char * logstr,
                       TSimpleCert ** myPointerToAMySimpleStack)

{
    // Try...
    static TSimpleCert      *  MySimpleStack = NULL;


    FILE             *  tmpFile     = NULL;
    STACK_OF(X509)   *  dupChain = NULL;
//    TSimpleCert      *  MySimpleStack= NULL;
    int                 certCount = 0;



#ifdef TESTBIO
    int err;
    int res = 0;
    BIO *in = NULL;
    X509 *x = NULL;
    STACK_OF(X509) *chain = NULL;
#else
    /* int                 errNo       = 0; */
    gss_cred_id_t    *  pcred       = NULL;
    gss_cred_id_t       cred        = GSS_C_NO_CREDENTIAL;
    X509             *  px509_cred  = NULL;
    STACK_OF(X509)   *  px509_chain = NULL;
#endif


    
    MySimpleStack = NULL;


    /*
     *
     * Inserting a user's certificate chain when needed (if it's not there yet... insert it))
     *
     */

#ifdef TESTBIO
#else
    /* Fetch user gss credential */
    if ( ( pcred = (gss_cred_id_t *) lcmaps_getArgValue("user_cred", "gss_cred_id_t", argc, argv) ) )
    {
        lcmaps_log_debug(2,"%s: address user_cred: %p\n", logstr,pcred);
        cred = *pcred;
        if (cred == GSS_C_NO_CREDENTIAL)
        {
            lcmaps_log(0,"%s: user gss credential is empty.\n", logstr);
            //goto fail_jobrep;
            return 0;
        }
    }
    else
    {
        lcmaps_log(0,"%s: could not get address of user credentials.\n", logstr);
        //goto fail_jobrep;
        return 0;
    }
#endif

    /* Testing */
#undef EXPORT_CREDENTIAL
#if EXPORT_CREDENTIAL
    if (cred)
    {
        gss_buffer_desc                deleg_proxy_filename;
        OM_uint32         major_status = 0;
        OM_uint32         minor_status = 0;

        major_status = gss_export_cred(&minor_status,
                                       cred,
                                       NULL,
                                       1,
                                       &deleg_proxy_filename);

        if (major_status == GSS_S_COMPLETE)
        {
            char *                     cp;

            lcmaps_log_debug(2,"%s: deleg_proxy_filename.value: %s\n", logstr,
                               deleg_proxy_filename.value);
            cp = strchr((char *)deleg_proxy_filename.value, '=');
            *cp = '\0';
            cp++;
            setenv((char *)deleg_proxy_filename.value, cp, 1);
            free(deleg_proxy_filename.value);
        }
        else
        {
            char *                      error_str = NULL;
            globus_object_t *           error_obj;

            error_obj = globus_error_get((globus_result_t) minor_status);

            error_str = globus_error_print_chain(error_obj);
            lcmaps_log (0,"%s: Error, gss_export_cred(): %s\n", logstr,error_str);

            //goto fail_jobrep;
            return 0;
        }
    }
#endif /* EXPORT_CREDENTIAL */



#ifdef TESTBIO
#else
    /*
     * Retrieve a newly created X509 struct and X509 chain from gss credential (should be freed)
     */
    if ( ( px509_cred = (X509 *) lcmaps_cred_to_x509(cred) ) )
    {
        lcmaps_log_debug(2,"%s: found X509 struct inside gss credential\n", logstr);
        lcmaps_log_debug(5,"%s: just for kicks: X509->name %s\n", logstr,px509_cred->name);
    }
    else
    {
        lcmaps_log(0,"%s: could not get X509 cred.\n", logstr);
    //    goto fail_jobrep;
    }
    if ( ( px509_chain = (STACK_OF(X509) *) lcmaps_cred_to_x509_chain(cred) ) )
    {
        lcmaps_log_debug(1,"%s: found X509 chain inside gss credential\n", logstr);
    }
    else
    {
        lcmaps_log(0,"%s: could not get X509 chain from the credentials.\n", logstr);
    //    goto fail_jobrep;
    }
#endif




/* VOMS stuff that I've lost because the Framework can parse it itself.  */




#ifdef TESTBIO
    in = BIO_new(BIO_s_file());
//    chain = load_chain("/home/gridtest/cvs/fabric_mgt/gridification/lcmaps/modules/voms/x509up_u500");

    lcmaps_log_debug (2, "%s: load chain...\n", logstr);
    chain = load_chain(getenv("X509_USER_PROXY"));
    lcmaps_log_debug (2, "%s: chain loaded\n", logstr);
    if (in)
    {
        if (BIO_read_filename(in, getenv("X509_USER_PROXY")) > 0)
        {

            lcmaps_log(2, "%s: found BIO file\n", logstr);
        }
        else
        {
            lcmaps_log(2, "%s: did not found BIO file!\n", logstr);
        }

//            x = PEM_read_bio_X509(in, NULL, 0, NULL);

//            res = VOMS_Retrieve(x, chain, RECURSE_CHAIN, vd, &err);

//            if (res)
//                print_vomsdata(vd);
//            else
//                lcmaps_log(2, "%s: VOMS Retrieve didn't found VOMS information.\n", logstr);
//        }
    }
    else
    {
        lcmaps_log(1, "Error: No BIO File!\n");
        goto fail_jobrep;
    }

//    if (!res)
//    {
//        lcmaps_log_debug(1, "%s: err: %s\n", logstr, retmsg[err]);
//    }

    // Check is the chain is there
    if (chain != NULL)
#else

    // Check if this chain is there
    if (px509_chain != NULL)
#endif
    {
        int ctCnt = 0;

        lcmaps_log_debug(2, "%s: The certificate chain is usable.\n", logstr);


#ifdef TESTBIO
        dupChain = (STACK_OF(type) *) sk_X509_dup(chain);
#else
        dupChain = (STACK_OF(type) *) sk_X509_dup(px509_chain);
#endif





        certCount = sk_X509_num(dupChain);
        lcmaps_log_debug (1, "%s: The chain consist of %d certificates.\n", logstr, certCount);

        MySimpleStack = malloc (sizeof(TSimpleCert) * (certCount + 1));
        for (ctCnt = 0; ctCnt < certCount; ctCnt++)
        {
            MySimpleStack[ctCnt].X509cert = sk_X509_pop(dupChain);
            if (MySimpleStack[ctCnt].X509cert != NULL)
            {
                char          buffer[1000];
                int           kar    = '\0';
                int           offset = 0;
                int           certSize = 100000;
                int           i = 0;

                // Make a (char *) out of a certificate
                tmpFile = tmpfile();

                X509_print_fp(tmpFile, MySimpleStack[ctCnt].X509cert);
                rewind(tmpFile);


                MySimpleStack[ctCnt].certStr = malloc(sizeof(char) * certSize);
                for (i = 0; i < certSize; i++) MySimpleStack[ctCnt].certStr[i] = '\0';
                for (i = 0; (kar = fgetc(tmpFile)) != EOF; i++)
                {
//                    // VERY UGLY BUG FIX !!!
//                    if (kar == '\'')   /* a '\'' = decimal 39 = hex 27 = oct 047 = html &#39; => Using Hex encoding */
//                    {
//                        MySimpleStack[ctCnt].certStr[i + offset]   = '%';
//                        MySimpleStack[ctCnt].certStr[i + ++offset] = '2';
//                        MySimpleStack[ctCnt].certStr[i + ++offset] = '7';
//                        lcmaps_log_debug(2, "%s: \tWarning: char \' encoded into %c27\n\t\t\t\t\t\t\t\tbecause of SQL problems. (A workaround needs to be developed)\n", logstr, '%');

                    // VERY Nice Bug Fix - this will alter a ' into a \' in the query, which is the perfect solution for this problem.
                    if (kar == '\'')
                    {
                        // Very easy and sweet fix
                        MySimpleStack[ctCnt].certStr[i + offset]   = '\\';
                        MySimpleStack[ctCnt].certStr[i + ++offset] = '\'';
                    }
                    else
                    {
                        MySimpleStack[ctCnt].certStr[i + offset] = kar;
                    }
                }

//                while ((kar = fgetc(tmpFile)) != EOF)
//                {
//                    if (certStr == NULL)
//                    {
//                        certStr = malloc(sizeof(char));
//                        certStr[0] = '\0';
//                    }
//                    certStr = realloc(certStr, (sizeof(char) * strlen(certStr)) + 2);
//                    sprintf(certStr, "%s%c", certStr, ch);
//
//                    certStr[i] = kar;
//                    i++;
//                }

                //MySimpleStack[ctCnt].certStr = strdup(certStr);

                X509_NAME_oneline(X509_get_subject_name(MySimpleStack[ctCnt].X509cert), buffer, 999);
                MySimpleStack[ctCnt].subject_name = malloc(sizeof(char) * 1000);
//                strncpy(MySimpleStack[ctCnt].subject_name, buffer, strlen(buffer));
                strcpy(MySimpleStack[ctCnt].subject_name, buffer);

                X509_NAME_oneline(X509_get_issuer_name(MySimpleStack[ctCnt].X509cert), buffer, 999);
                MySimpleStack[ctCnt].issuer_name = malloc(sizeof(char) * 1000);
//                strncpy(MySimpleStack[ctCnt].issuer_name, buffer, strlen(buffer));
                strcpy(MySimpleStack[ctCnt].issuer_name, buffer);

                MySimpleStack[ctCnt].strNotBefore = malloc(sizeof(char) * 20);
                MySimpleStack[ctCnt].strNotAfter  = malloc(sizeof(char) * 20);
                strncpy(MySimpleStack[ctCnt].strNotBefore,
                            sttm_2_char (ASN1_TIME_2_time_t(X509_get_notBefore(MySimpleStack[ctCnt].X509cert))), 20);
                strncpy(MySimpleStack[ctCnt].strNotAfter,
                            sttm_2_char (ASN1_TIME_2_time_t(X509_get_notAfter(MySimpleStack[ctCnt].X509cert))),  20);



                //////
                //
                //  X509_NAME * X509_get_issuer_name(X509 *a);
                //  X509_NAME * X509_get_subject_name(X509 *a);
                //
                //              X509_get_version(x) ASN1_INTEGER_get((x)->cert_info->version)
                //  /*          X509_get_serialNumber(x) ((x)->cert_info->serialNumber) */
                //  ASN1_TIME   X509_get_notBefore(x) ((x)->cert_info->validity->notBefore)
                //  ASN1_TIME   X509_get_notAfter(x) ((x)->cert_info->validity->notAfter)
                //              X509_extract_key(x)     X509_get_pubkey(x) /*****/
                //              X509_REQ_get_version(x) ASN1_INTEGER_get((x)->req_info->version)
                //  X509_NAME * X509_REQ_get_subject_name(x) ((x)->req_info->subject)
                //              X509_REQ_extract_key(a) X509_REQ_get_pubkey(a)
                //
                //              X509_CRL_get_version(x) ASN1_INTEGER_get((x)->crl->version)
                //  ASN1_TIME   X509_CRL_get_lastUpdate(x) ((x)->crl->lastUpdate)
                //  ASN1_TIME   X509_CRL_get_nextUpdate(x) ((x)->crl->nextUpdate)
                //  X509_NAME * X509_CRL_get_issuer(x) ((x)->crl->issuer)
                //              X509_CRL_get_REVOKED(x) ((x)->crl->revoked)
                //
                //
                //
                //
                //  Put here the code that will make a user_certificate
                //
                //  Select existance in user_certificates
                //  if exists check parent ... if same parent ... do nothing
                //  else insert it with the parent_cert
                //
                //  update jobs
                //  update users
                //
                ///////
            }
        }
    }


#ifdef TESTBIO
    if (chain) sk_X509_free(chain);
#endif


    if (dupChain) sk_X509_free(dupChain);
    if (tmpFile) fclose(tmpFile);

#ifdef TESTBIO
    if (x) X509_free(x);
    if (in) BIO_free(in);
#endif

    *myPointerToAMySimpleStack = MySimpleStack;

    return certCount;
}





/******************************************************************************
Function:   strMangleMySQL
Description:
         This function mangles a string to conform to be a MySQL understood 
         string.
         Basicly it means to encode the ' into a \' and such...
Parameters:
         char * strMangle

Returns:
    !NULL              : succes
    NULL               : failure
******************************************************************************/

char * strMangleMySQL (char * strMangle)
{
    if ( (strMangle) && (strlen(strMangle) != 0) )
    {
        int    i      = 0;
        int    offset = 0;
        char * tmpStr = NULL;
 
        tmpStr = malloc(sizeof(char) * strlen(strMangle) * 2);

        for (i = 0; i < strlen(strMangle); i++)
        {
            // Nice fix
            if (strMangle[i] == '\'')
            {
                tmpStr[i + offset]   = '\\';
                tmpStr[i + ++offset] = '\'';
            }
            else
            {
                tmpStr[i + offset] = strMangle[i];
            }
        }
        tmpStr[i + offset + 1] = '\0';

        return tmpStr;
    }
    else 
    { 
        return NULL;
    }
}


    /* End of uglyness */


/******************************************************************************
Function:   tripletLine2FQAN
Description:
         This function extracts the triplet information from the tripletLine 
         string and returns a FQAN. The tripletLine is in the same format as
         the vomapfile (or grid-mapfile) and the groupmapfile.
         This format is also supported with voms2gacl to get the VOMS info
         in a GACL file. Here it functions to convert the compare the different
         formats to the (standard) FQAN.

Parameters:
         tripletLine
 
Returns:
    !NULL              : succes
    NULL               : failure
******************************************************************************/

char * tripletLine2FQAN (char * tripletLine)
{
    char * pTL   = tripletLine;
    char   vo    [100];
    char   group [100];
    char   role  [100];
    char   cap   [100];
    int    i     = 0;
    char * search = NULL;
    char * logstr = "\tlcmaps_plugin_jobrepository-tripletLine2FQAN()";

    for (i = 0; i < 100; i++) { vo[i] = group[i] = role[i] = cap[i] = '\0'; }


//    while (strlen(pTL) > 0)  - Invalid pointer location: Going one step to far; which is then a pointer to boldly go where no pointer has gone before.
//    while (strlen(pTL) > 1)
    while ( ((pTL-tripletLine) < (strlen(tripletLine)+1)) && strlen(pTL) > 0)
    {
        i = 0;
        if (strncmp(pTL, "/VO=", strlen("/VO=")) == 0)
        {
            // Shift prefix
            pTL += strlen("/VO=");

            // Search for next prefix, if found then copy the offset
            search = pTL;
            while (strlen(search) > 0)
            {
                if ((strncmp(search, "/VO=", strlen("/VO=")) == 0) || 
                    (strncmp(search, "/GROUP=", strlen("/GROUP=")) == 0) || 
                    (strncmp(search, "/ROLE=", strlen("/ROLE=")) == 0)   ||
                    (strncmp(search, "/CAPABILITY=", strlen("/CAPABILITY=")) == 0))
                {
                    break;
                }
                i++;
                search++;
            }

            if (i < 100)
            {
	            strncpy(vo, pTL, i);
                pTL += strlen(vo);
            }
			else
			{
				lcmaps_log(0, "%s: Buffer overrun for \'vo\'\n", logstr);
			    return NULL;
			}
        }

        i = 0;

        if (strncmp(pTL, "/GROUP=", strlen("/GROUP=")) == 0)
        {
            pTL += strlen("/GROUP=");

            // Search for next prefix, if found then copy the offset
            search = pTL;
            while (strlen(search) > 0)
            {
                if ((strncmp(search, "/VO=", strlen("/VO=")) == 0) ||
                    (strncmp(search, "/GROUP=", strlen("/GROUP=")) == 0) ||
                    (strncmp(search, "/ROLE=", strlen("/ROLE=")) == 0)   ||
                    (strncmp(search, "/CAPABILITY=", strlen("/CAPABILITY=")) == 0))
                {
                    break;
                }

                i++;
                search++;
            }
 
            if (i < 100)
            {
                strncpy(group, pTL, i);
                pTL += strlen(group);
            }
			else
			{
				lcmaps_log(0, "%s: Buffer overrun for \'group\'\n", logstr);
			    return NULL;
			}
        }

        i = 0;
        if (strncmp(pTL, "/ROLE=", strlen("/ROLE=")) == 0)
        {
            pTL += strlen("/ROLE=");

            // Search for next prefix, if found then copy the offset
            search = pTL;
            while (strlen(search) > 0)
            {
                if ((strncmp(search, "/VO=", strlen("/VO=")) == 0) ||
                    (strncmp(search, "/GROUP=", strlen("/GROUP=")) == 0) ||
                    (strncmp(search, "/ROLE=", strlen("/ROLE=")) == 0)   ||
                    (strncmp(search, "/CAPABILITY=", strlen("/CAPABILITY=")) == 0))
                {
                    break;
                }

                i++;
                search++;
            }

            if (i < 100)
            {
                strncpy(role, pTL, i);
                pTL +=  strlen(role) - 1;
            }
			else
			{
				lcmaps_log(0, "%s: Buffer overrun for \'role\'\n", logstr);
			    return NULL;
			}
        }

        i = 0;
        if (strncmp(pTL, "/CAPABILITY=", strlen("/CAPABILITY=")) == 0)
        {
            pTL += strlen("/CAPABILITY=");

            // Search for next prefix, if found then copy the offset
            search = pTL;
            while (strlen(search) > 0)
            {
                if ((strncmp(search, "/VO=", strlen("/VO=")) == 0) ||
                    (strncmp(search, "/GROUP=", strlen("/GROUP=")) == 0) ||
                    (strncmp(search, "/ROLE=", strlen("/ROLE=")) == 0)   ||
                    (strncmp(search, "/CAPABILITY=", strlen("/CAPABILITY=")) == 0))
                {
                    break;
                }
                i++;
                search++;
            }
 
            if (i < 100)
            {
                strncpy(cap, pTL, i);
                pTL +=  strlen(cap) - 1;
            }
			else
			{
				lcmaps_log(0, "%s: Buffer overrun for \'capability\'\n", logstr);
			    return NULL;
			}
        }

        pTL++;
    }
    lcmaps_log_debug(5, " >> vo: %s (%d)  group: %s (%d)  role: %s (%d)  cap: %s (%d)  <<\n", vo, strlen(vo), group, strlen(group), role, strlen(role), cap, strlen(cap));

    return createFQAN (vo, group, role, cap);
}



/******************************************************************************
Function:   createFQAN
Description:
            Builds a FQAN of the vo, group, role and cap (capability) parameters
Parameters:
            char * vo, group, role, cap
Returns:
    !NULL              : succes
    NULL               : failure
******************************************************************************/ 
char * createFQAN(char * vo, char * group, char * role, char * cap)
{
    static char * fqan = NULL;
    int    i    = 0;

    if ((group == NULL) || (strlen(group) < 1))
        return NULL;


    // EVIL BUG FIX! - Because of " " in front of the Role
    if ((role != NULL) && (strlen(role) > 1) && (role[0] == ' ')) {  role++;  }

 
    fqan = malloc (250 * sizeof(char));
    for (i = 0; i < 250; i++)  { fqan[i] = '\0'; }

    // FQAN building ...
    sprintf(fqan, "%s", group);
    
    if ((role != NULL) && (strcmp(role, "NULL") != 0) && (strlen(role) > 0))
        sprintf(fqan, "%s/Role=%s", fqan, role);

    if ((cap != NULL)  && (strcmp(cap, "NULL") != 0)  && (strlen(cap) > 0)) 
        sprintf(fqan, "%s/Capability=%s", fqan, cap);

    return fqan;
}


#ifdef TESTBIO
static STACK_OF(X509) *load_chain(char *certfile)
{
    STACK_OF(X509_INFO) *sk=NULL;
    STACK_OF(X509) *stack=NULL, *ret=NULL;
    BIO *in=NULL;
    X509_INFO *xi;
    int first = 1;
    char * logstr = "load_chain";

    if(!(stack = sk_X509_new_null()))
    {
        lcmaps_log(0, "%s: memory allocation failure\n", logstr);
        goto end;
    }

    if(!(in=BIO_new_file(certfile, "r")))
    {
        lcmaps_log(0, "%s: error opening the file, %s\n", logstr,certfile);
        goto end;
    }

    /* This loads from a file, a stack of x509/crl/pkey sets */
    if(!(sk=PEM_X509_INFO_read_bio(in,NULL,NULL,NULL)))
    {
        lcmaps_log(0, "%s: error reading the file, %s\n", logstr,certfile);
        goto end;
    }

    /* scan over it and pull out the certs */
    while (sk_X509_INFO_num(sk))
    {
        /* skip first cert */
        if (first)
        {
            first = 0;
            continue;
        }
        xi=sk_X509_INFO_shift(sk);
        if (xi->x509 != NULL)
        {
            sk_X509_push(stack,xi->x509);
            xi->x509=NULL;
        }
        X509_INFO_free(xi);
    }
    if(!sk_X509_num(stack))
    {
        lcmaps_log(0, "%s: no certificates in file, %s\n", logstr,certfile);
        sk_X509_free(stack);
        goto end;
    }
    ret=stack;
end:
    BIO_free(in);
    sk_X509_INFO_free(sk);
    return(ret);
}
#endif

/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(
        int argc,
        char ** argv
)
{
    char * logstr = "\tlcmaps_plugin_jobrepository-plugin_initialize()";
    int i;

    // connStr = (char *) malloc (sizeof(char) * 10000);
    // for (i = 0; i < 1000; i++) connStr[i] = '\0';
 
    lcmaps_log_debug(1,"%s: passed arguments:\n", logstr);
    for (i=0; i < argc; i++)
    {
       lcmaps_log_debug(2,"%s: arg %d is %s\n", logstr, i, argv[i]);
    }
 
    /*
     * CERTDIR = The directory which contains the CA certificates
     * VOMSDIR = The directory which contains the certificates of the VOMS servers 
     */
 
    /*
     * Parse arguments, argv[0] = name of plugin, so start with i = 1
     */
    for (i = 1; i < argc; i++)
    {
/*
        if ( ((strcmp(argv[i], "-vomsdir") == 0) ||
              (strcmp(argv[i], "-VOMSDIR") == 0))
             && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 vomsdir = strdup(argv[i + 1]);
            }
            i++;
        }
        else if ( ((strcmp(argv[i], "-certdir") == 0) ||
                   (strcmp(argv[i], "-CERTDIR") == 0))
                   && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 certdir = strdup(argv[i + 1]);
            }
            i++;
        }
*/

//        else 
             if ( ((strcmp(argv[i], "-test") == 0) ||
                   (strcmp(argv[i], "-TEST") == 0)) )
        {
            test_jobrep = 1;
        }

        else if ( ( (strcmp(argv[i], "-JR_CONFIG") == 0) 
                    || 
                      (strcmp(argv[i], "-jr_config") == 0) ) 
                  && 
                    (i + 1 < argc) )
        {
            jr_config_file = strdup (argv[i + 1]);
            lcmaps_log (1, "%s: \tJob Repository Configuration file is: %s\n", logstr, jr_config_file);

            i++;
        }

/*
            lcmaps_get_jobrep_config(argv[i + 1], &jobrep_config);

            if ((jobrep_config) && (strlen(jobrep_config) != 0))
            {
                if (strlen(connStr) == 0)
                    strncpy(connStr, jobrep_config, strlen(jobrep_config));
                else
                    sprintf(connStr, "%s;%s", connStr, jobrep_config);
            }
            else
            {
                lcmaps_log(0,"%s: Error in initialization parameter: %s (failure); The file %s appears to be not available or empty.\n", logstr, argv[i], argv[i + 1]);
                return LCMAPS_MOD_FAIL;
            }
            i++;

        }
*/


        // Due to security compromise reasons, this should not be possible to be connecting with:
/*
        else if ( 
                  (strncmp(argv[i], "DSN=",      4) == 0) ||
                  (strncmp(argv[i], "HOST=",     5) == 0) ||
                  (strncmp(argv[i], "USER=",     5) == 0) ||
                  (strncmp(argv[i], "PASS=",     5) == 0) ||
                  (strncmp(argv[i], "DRIVER=",   7) == 0) ||
                  (strncmp(argv[i], "DATABASE=", 9) == 0) ||
                  (strncmp(argv[i], "PORT=",     5) == 0) ||
                  (strncmp(argv[i], "PASSWORD=", 9) == 0) ||
                  (strncmp(argv[i], "SERVER=",   7) == 0) ||
                  (strncmp(argv[i], "driver=",   7) == 0) ||
                  (strncmp(argv[i], "dsn=",      4) == 0) ||
                  (strncmp(argv[i], "host=",     5) == 0) ||
                  (strncmp(argv[i], "user=",     5) == 0) ||
                  (strncmp(argv[i], "port=",     5) == 0) ||
                  (strncmp(argv[i], "database=", 9) == 0) ||
                  (strncmp(argv[i], "server=",   7) == 0) ||
                  (strncmp(argv[i], "port=",     5) == 0) ||
                  (strncmp(argv[i], "password=", 9) == 0) ||
                  (strncmp(argv[i], "pass=",     5) == 0) )
        {
            lcmaps_log_debug(2, "------------------------------------->        %s\n", argv[i]);

            if (strlen(connStr) == 0)
                strncpy(connStr, argv[i], strlen(argv[i]));
            else
                sprintf(connStr, "%s;%s", connStr, argv[i]);
        }
*/

        else
        {
            lcmaps_log(0,"%s: Error in initialization parameter: %s (failure)\n", logstr, argv[i]);
            return LCMAPS_MOD_FAIL;
        }
    }

    lcmaps_log_debug(1,"%s: Initialization succeeded\n", logstr);
    return LCMAPS_MOD_SUCCESS;
} 


/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(
        int * argc,
        lcmaps_argument_t ** argv
)
{
    char * logstr = "\tlcmaps_plugin_voms-plugin_introspect()";
    static lcmaps_argument_t argList[] = {
        { "user_dn"     , "char *"           ,  1, NULL},
        { "user_cred"   , "gss_cred_id_t"    ,  0, NULL},
        { "job_request" , "char *"           ,  0, NULL},
        { NULL          , NULL               , -1, NULL}
    };

    lcmaps_log_debug(2,"%s: introspecting\n", logstr);

    *argv = argList;
    lcmaps_log_debug(2,"%s: before lcmaps_cntArgs()\n", logstr);
    *argc = lcmaps_cntArgs(argList);
    lcmaps_log_debug(2,"%s: address first argument: 0x%x\n", logstr,argList);

    lcmaps_log_debug(1,"%s: Introspect succeeded\n", logstr);
    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(
        int argc,
        lcmaps_argument_t * argv
)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}

/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(
        int argc,
        lcmaps_argument_t * argv
)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(
        int argc,
        lcmaps_argument_t * argv,
        int lcmaps_mode
)
{
    char             *  logstr      = "\tlcmaps_plugin_jobrepository-plugin_run()";
    int                 tryCnt      = 0;
    int                 connected   = 0;
//    FILE             *  tmpFile     = NULL;
    char             *  dn          = NULL;
    char             *  rsl         = NULL; 
    char             *  tmpRSL      = NULL;

//    struct vomsdata  *  vd          = NULL;
//    struct voms      ** vo          = NULL;
//    struct voms      *  v           = NULL;

    uid_t            *  uid         = NULL;
    int                 cntUid = 0;
    gid_t            *  priGid      = NULL;
    int                 cntPriGid = 0;
    gid_t            *  secGid      = NULL;
    int                 cntSecGid = 0;
    struct passwd    *  user_info   = NULL;
    struct group     *  group_info  = NULL;

    char             *  datetime = NULL;
    time_t              clock;
    struct tm        *  tmpTime;

    TResultSet       *  resultSet = NULL;
    char                query[1000000];
    char             *  user_id = NULL;
    char             *  credential_id = NULL;
    char             *  credential_group_id = NULL;
    char             *  cert_id = NULL;

    char             *  service_action_id = NULL;
    char             *  action_type_id = NULL;
    char             *  fqan_id = NULL;
    char             *  voms_attribute_id = NULL;
    char             *  jobrep_job_id = NULL;


// special to solve the UID/GIDs problem with and without VOMS
    TUnixUid            unixUid;
    TUnixGid         *  unixGids = NULL;
    int                 unixGidCount = 0;




//    STACK_OF(X509)   *  dupChain = NULL;
    TSimpleCert      *  MySimpleStack= NULL;
    int                 certCount = 0;

    char **             vo_cred_string_list = NULL;
    int                 cnt_vo_cred_string  = 0;

    int                 i = 0;
    int                 j = 0; 
//    int                 k = 0;

    /*
     * The beginning
     */
    if (lcmaps_mode == PLUGIN_RUN)
        logstr = "\tlcmaps_plugin_jobrepository-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = "\tlcmaps_plugin_jobrepository-plugin_verify()";
    else
    {
        lcmaps_log(0,
            "\tlcmaps_plugin_voms_jobrepository-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n",
            lcmaps_mode);
        goto fail_jobrep;
    }
    lcmaps_log_debug(1,"%s:\n", logstr);

    // Array of structs with vo_data_string, groupname, gid
//    lcmaps_vo_mapping_t ** vo_mapping = NULL;

    lcmaps_vo_mapping_t * vo_mapping = NULL;
    int                    cnt_vo_mapping = 0;

    

    /*
     * Before time ... there were files
     */

    //   Read file here and parse it

    if ( lcmaps_get_jobrep_config ( jr_config_file ) )
    {
        lcmaps_log(0,
            "\tlcmaps_plugin_voms_jobrepository-plugin_run_or_verify(): failure in opening or parsing the Job Repository Configuration File: %s\n",
            jr_config_file);
        goto fail_jobrep;
    }


    /*
     * In the beginning... there was time
     */

    time(&clock);
    tmpTime = gmtime(&clock);
    datetime = malloc(sizeof(char) * 20);
    snprintf(datetime, 20, "%04d-%02d-%02d %02d:%02d:%02d",
           tmpTime->tm_year + 1900, tmpTime->tm_mon + 1, tmpTime->tm_mday,
           tmpTime->tm_hour, tmpTime->tm_min, tmpTime->tm_sec);


    /* 
     * First execute a possibly very hard task: Converting the certificate chain in an array of TSimpleCert
     * That will increase the ease later when working with the SQL statements. 
     * Besides the cert itself (in a mangled char * for MySQL format), also some values on each cert are 
     * extracted from each cert and put into the struct of each array field
     */ 

/////////////////
    certCount = parseRawCertificates(argc, argv, logstr, &MySimpleStack);

    lcmaps_log_debug(1, "Amount of certs found in chain is: %d\n", certCount);
////////////////



//    // Get VO mapping
    vo_mapping          = getCredentialData(LCMAPS_VO_CRED_MAPPING, &cnt_vo_mapping);
    vo_cred_string_list = getCredentialData(LCMAPS_VO_CRED_STRING, &cnt_vo_cred_string);



   


    /*
     * Setting up Database connection
     */


    lcmaps_log_debug(1, "%s: \tTrying to connect to the DB...\n", logstr);


    /*
     * --- Test Mode ---
     */
    if (test_jobrep)
    {
        lcmaps_log      (0,"---- Job Repository enters Test Mode ----\n");

        testDatabaseODBC (jr_db_name, jr_db_user, jr_db_pwd);

        lcmaps_log      (0, "---- Job Repository Test Mode is finished, plugin failure is by design ----\n");
        goto fail_jobrep;
    }


    for (tryCnt = 0; (tryCnt < 5) && (!connected); tryCnt++)
    {
        
        //if (ODBC_Connect (connStr) != 0)
        //if (ODBC_Connect ("jobrepository", "root", "Lara-Navi") != 0)

        if (ODBC_Connect (jr_db_name, jr_db_user, jr_db_pwd) != 0)
        {
            lcmaps_log_debug(4, "%s: \t\tConnection errors from ODBC retrying...\n "); 
            ODBC_Errors ("ODBC_Connect, retry...");
            ODBC_Disconnect();
        }
        else
        {
            connected = 1;
            break;
        }
    }

    if (!connected)
    {
        lcmaps_log(0, "%s\n", ODBC_Errors ("ODBC_Connect"));
        ODBC_Disconnect ();
        goto fail_jobrep;
    }
    else
    {
        //Debug
        lcmaps_log_debug(1, "%s: Connected to the Job Repository Database\n", logstr); 
    }




    /*
     * Acquiring information
     */

    if ( (rsl = *(char **) lcmaps_getArgValue("job_request", "char *", argc, argv)) )
        lcmaps_log_debug(1,"%s: found job_request: %s\n", logstr, rsl);
    else
    {
        lcmaps_log (0,"%s: could not get value of job_request.\n", logstr);
        goto fail_jobrep;
    }

    if ( (dn = *(char **) lcmaps_getArgValue("user_dn", "char *", argc, argv)) )
        lcmaps_log_debug(1,"%s: found dn: %s\n", logstr,dn);
    else
    {
        lcmaps_log(0,"%s: could not get value of DN.\n", logstr);
        goto fail_jobrep;
    }
 
    uid    = getCredentialData(UID,     &cntUid);
    priGid = getCredentialData(PRI_GID, &cntPriGid);
    secGid = getCredentialData(SEC_GID, &cntSecGid);
 
    lcmaps_log_debug(1, "DN : %s\n", dn);




    /*
     * Insert ot select a user in the DB
     * Affected table: 'users'
     */


    // Select the user_id for the DN 
    sprintf(query, "select user_id from users where dn = '%s'", dn);



    // Debug    
    lcmaps_log_debug(2, "Query to execute is: '%s'\n", query);

    if (SQL_Query(query) != 0)
    {
        SQL_Rollback();

        lcmaps_log(0, "Rollback on query '%s'\n", query);
        goto fail_jobrep;
    }
    else
    {
        lcmaps_log_debug (5, "execution succeeded of last query\n");

        if (SQL_GetQueryResult(&resultSet) > 0)
        {
            lcmaps_log_debug (5, "Getting user_id\n");


            if (SQL_GetValue(resultSet, "user_id", &user_id) != 0)
            {
                lcmaps_log (0, "Could not get value: %s from the DB resultset with query: %s\n", "user_id", query);
                SQL_Rollback();
                goto fail_jobrep;
            }
        }
        else
        {
            // Probably a new user
            sprintf(query, "insert into users (dn) values ('%s')", dn);

            lcmaps_log_debug (5, "Executing insertion of new user with: %s\n", query);

            if (SQL_Query(query) != 0)
            {
                lcmaps_log (0, "Could not insert new user with DN: %s\n", dn);
                SQL_Rollback();
                goto fail_jobrep;
            }
            else
            {
                sprintf(query, "select last_insert_id()");
                if (SQL_Query(query) != 0)
                {
                    SQL_Rollback();
                    goto fail_jobrep;
                }
                if (SQL_GetQueryResult(&resultSet) > 0)
                {
                    if (SQL_GetValue(resultSet, "last_insert_id()", &user_id) != 0)
                    {
                        lcmaps_log (0, "Could not get value: %s from resultset with query: %s\n", "last_insert_id()", query);
                        SQL_Rollback();
                        goto fail_jobrep;
                    }
                }
                else
                {
                    lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
                    SQL_Rollback();
                    goto fail_jobrep;
                }
            }
        }
    }





    /*
     *  Insert of User Certificates
     *  Affected table: 'user_certificates'
     */

    for (i = 0; i < certCount; i++)
    {
        lcmaps_log_debug (2, "doing the user_certs, round %d\n", i);


        // Preparing the SQL string
        if ((cert_id != NULL) && (strlen(cert_id) > 0))
        {
            sprintf(query, 
                    "insert into user_certificates (user_id, subject, cert, first_use, valid_from, valid_until, parent_cert_id) values (%s, '%s', '%s', '%s', '%s', '%s', %s)",
                                    user_id,
                                    MySimpleStack[i].subject_name,
                                    MySimpleStack[i].certStr,
                                    datetime,
                                    MySimpleStack[i].strNotBefore,
                                    MySimpleStack[i].strNotAfter,
                                    cert_id);
        }
        else
        {
            sprintf(query, 
                    "insert into user_certificates (user_id, subject, cert, first_use, valid_from, valid_until) values (%s, '%s', '%s', '%s', '%s', '%s')",
                                    user_id,
                                    MySimpleStack[i].subject_name,
                                    MySimpleStack[i].certStr,
                                    datetime,
                                    MySimpleStack[i].strNotBefore,
                                    MySimpleStack[i].strNotAfter);
        }


        lcmaps_log_debug (2, "Query to execute is: %s\n", query);

        // Executing the SQL Query
        if (SQL_Query(query) != 0)
        {
            // Insert failed, we presume a record was already present, so let's catch the cert_id
            //                BUT if it is NOT selectable and this fails down here (because the searched record doesn't exist... then
            //                The insertion did really fail. For the moment this is caused by the ' character in the certificates signature(s).
            //                Which is a delimiter for strings that should be inserted into a varchar kinda field in a database.
            sprintf(query, "select cert_id from user_certificates where subject = '%s' and valid_from = '%s' and valid_until = '%s'", 
                                   MySimpleStack[i].subject_name, 
                                   MySimpleStack[i].strNotBefore,
                                   MySimpleStack[i].strNotAfter);

            lcmaps_log_debug (2, "Query to execute is: %s\n", query);
            if (SQL_Query(query) != 0)
            {
                SQL_Rollback();
                goto fail_jobrep;
            }

            if (SQL_GetQueryResult(&resultSet) > 0)
            {
                if (SQL_GetValue(resultSet, "cert_id", &cert_id) != 0)
                {
                    lcmaps_log(0, "Could not get value: %s from the DB resultset with query: %s\n", "cert_id", query);
                    SQL_Rollback();
                    goto fail_jobrep;
                }
            }
            else
            {
                lcmaps_log(0, "Empty resultset for query: %s\n", query);
                SQL_Rollback();
                goto fail_jobrep;
            }
        }
        else
        {
            // Insert succeeded
            sprintf(query, "select last_insert_id()");
            if (SQL_Query(query) != 0)
            {
                SQL_Rollback();
                goto fail_jobrep;
            }

            if (SQL_GetQueryResult(&resultSet) > 0)
            {
                if (SQL_GetValue(resultSet, "last_insert_id()", &cert_id) != 0)
                {
                    lcmaps_log(0, "Could not get value: %s from the resultset\n", "last_insert_id()");
                    SQL_Rollback();
                    goto fail_jobrep;
                }
            }
            else
            {
                lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
                SQL_Rollback();
                goto fail_jobrep;
            }
        }
    }


    /*
     * Enter the Service Action
     * The Service Action acts as a place holder for the action that has been executed
     * The string field 'action_type' is specific to each action
     * Currently the only 'action_type' available is LCMAPS. Plans are made to co-op with other 
     * services through different interfaces to add more data about a user's action on one site
     * Very handy for forensic data mining.
     *
     * Affected table: service_action, action_types
     */


    // Select the action_type first!
    sprintf(query, "select action_type_id from action_types where action_type = 'LCMAPS'");

    // Debug
    lcmaps_log_debug (2, "Query to execute: %s\n", query);

    if (SQL_Query(query) != 0)
    {
        SQL_Rollback();
        lcmaps_log_debug (0, "Rollback on query: %s\n", query);
        goto fail_jobrep;
    }
    if (SQL_GetQueryResult(&resultSet) > 0)
    {
        if (SQL_GetValue(resultSet, "action_type_id", &action_type_id) != 0)
        {
            lcmaps_log (0, "Could not get value: %s from resultset with query: %s\n", "action_type_id", query);
            SQL_Rollback();
            goto fail_jobrep;
        }
    }
    else
    {
        lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
        SQL_Rollback();
        goto fail_jobrep;
    }



    /*
     * Each service_action is unique and will be inserted into the database as a unique record
     * Perhaps there is no certificate chain available ... then matching must be done on the dates between the various records
     */

    // Query prepare
    if ( (cert_id != NULL) && (strlen(cert_id) > 0) ) 
    {
        sprintf(query, "insert into service_actions (use_time, user_id, cert_id, action_type_id) values ('%s', %s, %s, %s)", 
                       datetime, user_id, cert_id, action_type_id);
    }
    else
    {
        sprintf(query, "insert into service_actions (use_time, user_id, action_type_id) values ('%s', %s, %s)", 
                       datetime, user_id, action_type_id);
    }

    
    // Debug
    lcmaps_log_debug (2, "Query to execute: %s\n", query);

    // Query execute
    if (SQL_Query(query) != 0)
    {
        SQL_Rollback();
        goto fail_jobrep;
    }
    else
    {
        sprintf(query, "select last_insert_id()");
        if (SQL_Query(query) != 0)
        {
            SQL_Rollback();
            goto fail_jobrep;
        }
        if (SQL_GetQueryResult(&resultSet) > 0)
        {
            if (SQL_GetValue(resultSet, "last_insert_id()", &service_action_id) != 0)
            {
                lcmaps_log (0, "Could not get value: %s from resultset with query: %s\n", "last_insert_id()", query);
                SQL_Rollback();
                goto fail_jobrep;
            }
        }
        else
        {
            lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
            SQL_Rollback();
            goto fail_jobrep;
        }
    }

    
    /*
     *  The VOMS stuff is linked to the unix_uid & unix_gid tables, these last two tables are linked to the jobs table
     *  I need to know the Jobs table in advance, so I need to insert it into the database
     *  Affected table: 'jobs'
     */

    tmpRSL = strMangleMySQL (rsl);
    // sprintf(query, "insert into jobs (job_id, user_id, submit_cert_id, rsl, creation_time) values ('%s', %s, %s, '%s', '%s')", getenv("JOB_REPOSITORY_ID"), user_id, cert_id, tmpRSL, datetime);
    sprintf(query, "insert into jobs (job_repository_id, user_id, submit_cert_id, rsl, service_id) values ('%s', %s, %s, '%s', %s)", getenv("JOB_REPOSITORY_ID"), user_id, cert_id, tmpRSL, service_action_id);


    // Debug
    lcmaps_log_debug (2, "Query to execute: %s\n", query);

    if (SQL_Query(query) != 0)
    {
        lcmaps_log(0, "Error: possible problem is an already existing job_id: %s with query: %s\n",  getenv("JOB_REPOSITORY_ID"), query);
        SQL_Rollback();
        goto fail_jobrep;
    }
    else
    {
        sprintf(query, "select last_insert_id()");
        if (SQL_Query(query) != 0)
        {
            SQL_Rollback();
            goto fail_jobrep;
        }
        if (SQL_GetQueryResult(&resultSet) > 0)
        {
            if (SQL_GetValue(resultSet, "last_insert_id()", &jobrep_job_id) != 0)
            {
                lcmaps_log (0, "Could not get value: %s from resultset with query: %s\n", "last_insert_id()", query);
                SQL_Rollback();
                goto fail_jobrep;
            }
        }
        else
        {
            lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
            SQL_Rollback();
            goto fail_jobrep;
        }
    }




/*************************
 Depricated functionality
**************************/


    /*
     * Insert a status 'gatekeeper' next with the job record
     * Affected table: 'job_status'
     */

/*
    sprintf(query, "insert into jobstatus (job_id, status, status_change) values ('%s', '%s', '%s')",
            getenv("JOB_REPOSITORY_ID"), "LCMAPS", datetime);


    lcmaps_log_debug(0, "Query to execute: %s\n", query);

    if (SQL_Query(query) != 0)
    {
        lcmaps_log(0, "Error: could not execute query: %s\n", query);
        SQL_Rollback();
        goto fail_jobrep;
    }


*/




    /*
     * I need to get the uid information to put into the database
     */


    
/*********************************************************

     \            /    ------     |        |   ------
      \          /    /      \    |\      /|  /      
       \        /    |        |   | \    / |  \
        \      /     |        |   |  \  /  |   ------
         \    /      |        |   |   \/   |         \
          \  /        \      /    |        |         /
           \/          ------     |        |   -----

*********************************************************/

    /*
     * If there are no VOMS credentials then add the GID to the database without consolidating the FQAN->GID mapping
     * Note: this means that I'm deliberatly relying on the mapping ... which doesn't hold if the mechanisms change in the LCMAPS policies...
     */

//    unixUid 
//    unixGids
//    unixGidCount

    /*
     * I need to get the UID information to put into the database
     */

    if (cntUid > 0)
    {
        unixUid.uid = uid[0];
        if ((user_info = getpwuid(uid[0])) != NULL)
        {
            unixUid.uid_name = user_info->pw_name;
        }
        unixUid.job_id  = jobrep_job_id;
        unixUid.user_id = user_id;

        /* 
         * The voms_attribute_id value needs to be added to this struct when we are up to it. 
         */
    }

    

    /*
     * Do the same but for the GID(s) prior to inserting them into the DB
     */
    unixGids = malloc (sizeof (TUnixGid) * (cntSecGid + cntPriGid));

    if (cntPriGid > 0)
    {
        unixGids[0].gid      = priGid[0];
        if ((group_info = getgrgid (priGid[0])) != NULL)
        {
            unixGids[0].gid_name = group_info->gr_name;
        }
        unixGids[0].job_id  = jobrep_job_id;
        unixGids[0].user_id = user_id;
        unixGids[0].is_primary = 1;

        /* 
         * The voms_attribute_id value needs to be added to this struct when we are up to it. 
         */
    }

    if (cntSecGid > 0)
    {
        for (i = 0; i < (cntPriGid + cntSecGid); i++ )
        {
            unixGids[i + cntPriGid].gid      = secGid[i];
            if ((group_info = getgrgid (secGid[i])) != NULL)
            {
                unixGids[i + cntPriGid].gid_name = group_info->gr_name;
            }
            unixGids[i + cntPriGid].job_id  = jobrep_job_id;
            unixGids[i + cntPriGid].user_id = user_id;
            unixGids[i + cntPriGid].is_primary = 0;
        }

        /* 
         * The voms_attribute_id value needs to be added to this struct when we are up to it. 
         */
    }

    
    /* 
     * Info:
     * The three structures are complete, if there is VOMS info, this needs to be attached to the structures, to do the structure 2 database insertions.
     *
     */


    // VOMS FQANs
    for (i = 0; i < cnt_vo_mapping; i++)
    {
        /*
         *  First getting the FQAN's id in 'fqan_id'
         *  Affectes table: 'voms_fqans'
         */

        // Select existing record of the FQAN
        lcmaps_log_debug (4, "FQAN: %d  %s\n", (vo_mapping[i]).vostring, (vo_mapping[i]).vostring);

        sprintf(query, "select fqan_id from voms_fqans where fqan = '%s'", (vo_mapping[i]).vostring);

        lcmaps_log_debug(2, "Query to execute: %s\n", query);

        if (SQL_Query(query) != 0)
        {
            lcmaps_log_debug (0, "Failure in query execution: %s\n", query);

            SQL_Rollback();
            goto fail_jobrep;
        }
        else
        {
            lcmaps_log_debug (2, "Succes in query execution: %s\n", query);

            if (SQL_GetQueryResult(&resultSet) > 0)
            {
                if (SQL_GetValue(resultSet, "fqan_id", &fqan_id) != 0)
                {
                    lcmaps_log (0, "Could not get value: %s from the DB resultset with query: %s\n", "fqan_id", query);
                    SQL_Rollback();
                    goto fail_jobrep;
                }
                else
                {
                    lcmaps_log_debug (2, "Got value: %s\n", fqan_id);
                }
            }
            else
            {
                // New FQAN found
                sprintf(query, "insert into voms_fqans (fqan) values ('%s')", (vo_mapping[i]).vostring);
                
                lcmaps_log_debug (2, "Query to execute: %s\n", query);

                if (SQL_Query(query) != 0)
                {
                    lcmaps_log (0, "Could not insert new FQAN with FQAN: %s\n", (vo_mapping[i]).vostring);
                    SQL_Rollback();
                    goto fail_jobrep;
                }
                else
                {
                    lcmaps_log_debug (4, "Succes in query execution: %s\n", query);

                    sprintf(query, "select last_insert_id()");
                    if (SQL_Query(query) != 0)
                    {
                        SQL_Rollback();
                        goto fail_jobrep;
                    }
                    else
                    {
                        if (SQL_GetQueryResult(&resultSet) > 0)
                        {
                            if (SQL_GetValue(resultSet, "last_insert_id()", &fqan_id) != 0)
                            {
                                lcmaps_log (0, "Could not get value: %s from resultset with query: %s\n", "last_insert_id()", query);
                                SQL_Rollback();
                                goto fail_jobrep;
                            }
                        }
                        else
                        {
                            lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
                            SQL_Rollback();
                            goto fail_jobrep;
                        }
                    }
                }
            }
        }



        /*
         *  Secondly getting the VOMS Server id in 'server_id'
         *  NOTE: This little feature is switched off - By means of non-implementation :-)
         *  Affects table: 'voms_server'
         */

        // Will be implemented later

        lcmaps_log_debug(4, "No voms server_id checking yet!\n");



        /*
         * Select existing record of the 'action_types'
         * Done for the voms_attributes
         */

        sprintf(query, "select action_type_id from action_types where action_type = '%s'", "LCMAPS");
        
        lcmaps_log_debug(2, "Query to execute: %s\n", query);

        if (SQL_Query(query) != 0)
        {
            SQL_Rollback();
            goto fail_jobrep;
        }
        else
        {
            if (SQL_GetQueryResult(&resultSet) > 0)
            {
                if (SQL_GetValue(resultSet, "action_type_id", &action_type_id) != 0)
                {
                    lcmaps_log (0, "Could not get value: %s from the DB resultset with query: %s\n", "action_type_id", query);
                    SQL_Rollback();
                    goto fail_jobrep;
                }
            }
            else
            {
                lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
                SQL_Rollback();
                goto fail_jobrep;
            }
        }





        /*
         *  Thirdly is getting it all together in the 'voms_attributes' table
         *  Affects table: 'voms_attributes'
         */

        
        // Query prepare

        // sprintf(query, "insert into voms_attributes (valid_from, valid_until, use_time, fqan_id, server_id, user_id, service_action_id) values ( )");
        sprintf(query, 
                "insert into voms_attributes (fqan_id, user_id, service_action_id) values (%s, %s, %s)", fqan_id, user_id, service_action_id);


        lcmaps_log_debug(2, "Query to execute: %s\n", query);

        // Query execute
        if (SQL_Query(query) != 0)
        {
            SQL_Rollback();
            goto fail_jobrep;
        }
        else
        {
            sprintf(query, "select last_insert_id()");
            if (SQL_Query(query) != 0)
            {
                SQL_Rollback();
                goto fail_jobrep;
            }
            if (SQL_GetQueryResult(&resultSet) > 0)
            {
                if (SQL_GetValue(resultSet, "last_insert_id()", &voms_attribute_id) != 0)
                {
                    lcmaps_log (0, "Could not get value: %s from resultset with query: %s\n", "last_insert_id()", query);
                    SQL_Rollback();
                    goto fail_jobrep;
                }
            }
            else
            {
                lcmaps_log(0, "ResultSet is empty for query: %s\n", query);
                SQL_Rollback();
                goto fail_jobrep;
            }
        }



        if (i == 0)
        {
            /* 
             * UnixUid link with first found voms_attribute_id
             * I can't get a better estimate to which it has been mapped, but this assumption is very solid though!
             */
            unixUid.voms_attribute_id = voms_attribute_id;
        }

        for (j = 0; j < (cntSecGid + cntPriGid); j++)
        {
            if (unixGids[j].gid = (vo_mapping[i]).gid)
            {
                unixGids[j].voms_attribute_id = voms_attribute_id;
                break;
            }
        }
    } /* End of VOMS stuff... but not for the Unix UID/GIDs */
    

    // VO <-> Group Mapping
    /*
     *  Affected tables: unix_uid & unix_gids
     */


    // unix UID
    if (unixUid.voms_attribute_id != NULL)
    {
        sprintf(query,
                    "insert into unix_uids (uid, uid_name, voms_attribute_id, job_id, user_id, use_time) values (%d, '%s', %s, %s, %s, '%s')",
                    unixUid.uid,
                    (unixUid.uid_name == NULL ? "''" : unixUid.uid_name),
                    unixUid.voms_attribute_id,
                    unixUid.job_id,
                    unixUid.user_id,
                    datetime);
    }
    else
    {
        sprintf(query,
                    "insert into unix_uids (uid, uid_name, job_id, user_id, use_time) values (%d, '%s', %s, %s, '%s')",
                    unixUid.uid,
                    (unixUid.uid_name == NULL ? "''" : unixUid.uid_name),
                    unixUid.job_id,
                    unixUid.user_id,
                    datetime);
    }



    // Query execute
    lcmaps_log_debug(2, "Query to execute: %s\n", query);
    if (SQL_Query(query) != 0)
    {
        SQL_Rollback();
        goto fail_jobrep;
    }    


    // unix GIDs
    for (j = 0; j < (cntSecGid + cntPriGid); j++)
    {
        // Insert Secondary Unix GID
        if (unixUid.voms_attribute_id != NULL)
        {
            sprintf(query, 
                     "insert into unix_gids (gid, gid_name, voms_attribute_id, job_id, user_id, is_primary, use_time) values (%d, '%s', %s, %s, %s, %d, '%s')",
                     unixGids[j].gid,
                     (unixGids[j].gid_name == NULL ? "''" : unixGids[j].gid_name),
                     unixGids[j].voms_attribute_id,
                     unixGids[j].job_id,
                     unixGids[j].user_id,
                     unixGids[j].is_primary,
                     datetime);
        }
        else
        {
            sprintf(query, 
                     "insert into unix_gids (gid, gid_name, job_id, user_id, is_primary, use_time) values (%d, '%s', %s, %s, %d, '%s')",
                     unixGids[j].gid,
                     (unixGids[j].gid_name == NULL ? "''" : unixGids[j].gid_name),
                     unixGids[j].job_id,
                     unixGids[j].user_id,
                     unixGids[j].is_primary,
                     datetime);
        }


        lcmaps_log_debug(2, "Query to execute: %s\n", query);

        // Query execute
        if (SQL_Query(query) != 0)
        {
            SQL_Rollback();
            goto fail_jobrep;
        }
    }


    /* succes */
// success_voms:

    lcmaps_log_debug(1,"%s: done\n", logstr);

#ifndef SPIT_SQL_ONLY
    SQL_Commit();
    ODBC_Disconnect();
#endif


    // Free stuff
//    if (dupChain) sk_X509_free(dupChain);
//    if (tmpFile) fclose(tmpFile);
#ifdef TESTBIO
    if (x) X509_free(x);
    if (in) BIO_free(in);
#endif

//    if (resultSet) TResultSet_free(resultSet);
    if (user_id) free(user_id);
    if (credential_id) free(credential_id);

    if (credential_group_id) free(credential_group_id);
    if (cert_id) free(cert_id);
    if (MySimpleStack) TSimpleCert_free(MySimpleStack, certCount);

    if (datetime) free(datetime);
//    if (px509_cred) X509_free(px509_cred);
//    if (px509_chain) sk_X509_free(px509_chain);


    lcmaps_log_debug(1, "done?\n");

    lcmaps_log_time(0,"%s: job repository plugin succeeded\n", logstr);

    return LCMAPS_MOD_SUCCESS;

fail_jobrep:

#ifndef SPIT_SQL_ONLY
    if (connected) SQL_Rollback();

    ODBC_Disconnect();
#endif

    // Free stuff

    lcmaps_log_debug(1,"%s: doing VOMS_Destroy\n", logstr);
//    VOMS_Destroy(vd);
    lcmaps_log_debug(1,"%s: done\n", logstr);

    // Free stuff

/**************
    if (dupChain) sk_X509_free(dupChain);
    if (tmpFile) fclose(tmpFile);
#ifdef TESTBIO
    if (x) X509_free(x);
    if (in) BIO_free(in);
#endif
***************/


//    if (resultSet) TResultSet_free(resultSet);
    if (user_id) free(user_id);
    if (credential_id) free(credential_id);
 
    if (credential_group_id) free(credential_group_id);
    if (cert_id) free(cert_id);
    if (MySimpleStack) TSimpleCert_free(MySimpleStack, certCount);
 
    if (datetime) free(datetime);
//    if (px509_cred) X509_free(px509_cred);
//    if (px509_chain) sk_X509_free(px509_chain);

    lcmaps_log_time(0,"%s: job repositroy plugin failed\n", logstr);

    return LCMAPS_MOD_FAIL;
}



/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
    char * logstr = "\tlcmaps_plugin_voms-plugin_terminate()";
    lcmaps_log_debug(1,"%s: terminating\n", logstr);
//    if (vomsdir) free(vomsdir);
//    if (certdir) free(certdir);
//    if (connStr) free(connStr);

    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   print_vomsdata
Description:
    prints vomsdata from a 'struct vomsdata *'
Parameters:
    struct vomsdata *
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
/*
void print_vomsdata(struct vomsdata *d)
{
    char * logstr = "\tlcmaps_plugin_voms-print_vomsdata()";
    struct voms **vo = d->data;
    struct voms *v;
    int k = 0;
    int j =0;
  
    while(vo[k])
    {
        v = vo[k++];
        lcmaps_log_debug(2,"%s: %d *******************************************\n", logstr,k);
        lcmaps_log_debug(2,"%s: SIGLEN: %d\nUSER: %s\n", logstr, v->siglen, v->user);
        lcmaps_log_debug(2,"%s: UCA: %s\nSERVER: %s\n", logstr, v->userca, v->server);
        lcmaps_log_debug(2,"%s: SCA: %s\nVO: %s\n", logstr, v->serverca, v->voname);
        lcmaps_log_debug(2,"%s: URI: %s\n", logstr, v->uri);
//        fprintf(stderr,"%s: DATE1: %s\n", logstr, d2i_ASN1_TIME(v->date1));
//        fprintf(stderr,"%s: DATE2: %s\n", logstr, d2i_ASN1_TIME(v->date2));
//        lcmaps_log_debug(2,"%s: DATE1: %s\n", logstr, v->date1);
//        lcmaps_log_debug(2,"%s: DATE2: %s\n", logstr, v->date2);

        switch (v->type)
        {
        case TYPE_NODATA:
            lcmaps_log_debug(2,"%s: NO DATA\n", logstr);
            break;
        case TYPE_CUSTOM:
            lcmaps_log_debug(2,"%s: %*s\n", logstr, v->datalen - 10, v->custom);
            break;
        case TYPE_STD:
            j = 0;
            while (v->std[j])
            {
                lcmaps_log_debug(1,"%s: GROUP: %s\tROLE: %s\tCAP: %s\n", logstr,v->std[j]->group,
                     v->std[j]->role,v->std[j]->cap);
                j++;
            }
            break;
        }
    }

    if (d->workvo)
        lcmaps_log_debug(1,"%s: WORKVO: %s\n", logstr, d->workvo);

    if (d->extra_data)
        lcmaps_log_debug(1,"%s: EXTRA: %s\n", logstr, d->extra_data);
}
*/

/******************************************************************************
CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/lcmaps-plugins-jobrep/src/jobrep/lcmaps_jobrep.c,v $
    $Date: 2005-12-07 13:56:54 $
    $Revision: 1.17 $
    $Author: okoeroo $
******************************************************************************/


/******************************************************************************
Function:   ASN1_GENERALIZEDTIME_2ilb
Description:
        converts a ASN1_TIME into a time_t
        (It's a little hack, but should work nice and neat).

Parameters:
        a ASN1_TIME, 

Returns:
        a time_t

******************************************************************************/
static struct tm * ASN1_TIME_2_time_t(ASN1_TIME *asn1_time)
{
        //time_t MyTime = (time_t) 0;  
        int i;
        static struct tm newTimeSpace;

        if (asn1_time->length != 13) 
            return NULL;
            //return (time_t) -1;

        for (i = 0; i < asn1_time->length / 2; i++)
        {
            char tmp[3];
            
            tmp[0] = asn1_time->data[0 + (i*2)];
            tmp[1] = asn1_time->data[1 + (i*2)];
            tmp[2] = '\0';
            
            switch(i)
            {
                case 0 : if (atoi(tmp) > 69)
                              newTimeSpace.tm_year = atoi(tmp) + 1900;
                         else
                              newTimeSpace.tm_year = atoi(tmp) + 100;
                         break;
                case 1 : newTimeSpace.tm_mon  = atoi(tmp) - 1; break;
                case 2 : newTimeSpace.tm_mday = atoi(tmp); break;
                case 3 : newTimeSpace.tm_hour = atoi(tmp); break;
                case 4 : newTimeSpace.tm_min  = atoi(tmp); break;
                case 5 : newTimeSpace.tm_sec  = atoi(tmp); break;
            }
        }
        //MyTime = mktime(&newTimeSpace);
        //return MyTime;

        return (&newTimeSpace);
}


/******************************************************************************
Function:   sttm_2_char
Description:
        converts a struct tm into a humanly readable char * of the time it's
        presenting.
        The representation will be in a GMT time, not Local.
 
Parameters:
        time
 
Returns:
        The time in a humanly readle format (and compatible to MySQL) as known
        in the variations of GMT (GMT conversion of the time parameter seen as
        a localtime), LOCAL (plain old local time) or LEAVE_TIME_ITS_OK (this
        means you should not convert/change the time but just leave it, in
        reality this means give me the time as a localtime thus without
        conversion).
 
******************************************************************************/
char * sttm_2_char (const struct tm * MyDateTime)
{
    char  * datetime = malloc(sizeof(char) * 21);

    snprintf(datetime, 21, "%04d-%02d-%02d %02d:%02d:%02d",
             MyDateTime->tm_year + 1900, MyDateTime->tm_mon + 1, MyDateTime->tm_mday,
             MyDateTime->tm_hour, MyDateTime->tm_min, MyDateTime->tm_sec);
 
    return datetime;
}



/******************************************************************************
Function:    strnclean
Description: cleans a dirty string, forces '\0' on the bufsize of **s
Parameters:
     input  : char **s
     output : char **s (cleaned string and NULLified)
Returns:
    SUCCESS
    FAILURE
 
******************************************************************************/
int strnclean (char **s, int bufsize)
{
    int i = 0;
 
    if (s == NULL)
        return -1;
 
    if (*s == NULL)
        return -1; 
 
    for (i = 0; i < bufsize; i++)
    {
        (*s)[i] = '\0';
    }
 
    return 0;
}
 
 
/******************************************************************************
Function:    strclean
Description: cleans a dirty string, forces '\0' on the strlen() of **s
Parameters:
     input  : char **s
     output : char **s (cleaned string and NULLified)
Returns:
    SUCCESS
    FAILURE
 
******************************************************************************/
int strclean (char **s)
{
    return strnclean(s, strlen(*s));
}
/*****************************************************************************/


/******************************************************************************
Function:   lcmaps_get_jobrep_config()

Description:
    Get the Job Repository configuration string from file
    This contains the DNS of the server, the database name, username and password.

Parameters:
    path:   path to the jobrep_config file containing the configuration
    jobrep_config : variable to set the config string in.

Returns:
    0 on success.
    1 on failure
******************************************************************************/
/*!
    \fn lcmaps_get_jobrep_config(
        const char * path,
              char ** jobrep_config
    )
    \brief Get the Job Repository configuration from file

    This function tries to read the Job Repository configuration from the 
    jobrep_config file.
    It also tests if the access bits of the file are correctly set.


    \param path     the path to the jobrep_config file containing the config string.
    \param jobrep_config variable to set the configuration string in

    \retval 0 success
    \retval 1 other failure
*/

static int lcmaps_get_jobrep_config ( const char * path )
{
    char *        logstr = "\tlcmaps_plugin_jobrepository-lcmaps_get_jobrep_config()";
    FILE        * fp = NULL;
    struct stat   buf;
    int           c;
    int           count = 0;
    char        * tempconfig = NULL;

    lcmaps_log_debug(3,"%s: path: %s \n", logstr, path);

    /* get the stats of the file */
    if ((stat(path, &buf)) == -1)
    {
        lcmaps_log(0, "%s: unable to get stat of jobrep_config file: %s\n", logstr, path);
        return 1;
    }

    /* test if it's a file */
    if ( !(S_ISREG(buf.st_mode)))
    {
        lcmaps_log(0, "%s: jobrep_config file is not a file: %s\n", logstr, path);
        return 1;
    }

    /* test if root is owner */
/*
    if (buf.st_uid != 0) {
        lcmaps_log(0, "%s: jobrep_config file is not owned by root (readonly for root) : %s\n", logstr, path);
        return 1;
    }
*/

    /* test if file has the access bits set correctly */
    if ((0000777  & (buf.st_mode) ) != S_IRUSR )
    {
        lcmaps_log(0, "%s: jobrep_config file has incorrect accessibility (readonly for root) : %s\n", logstr, path);
        return 1;
    }

    /* try to open the file */
    if ((fp = fopen(path, "r")) == NULL)
    {
        lcmaps_log(0, "%s: unable to open jobrep_config file: %s\n", logstr, path);
        return 1;
    }
    /* read the passwd from the file */
    /* I thought tabs and spaces were not allowed in the passwd, but they are, so
     * restore original line
     */
//    while ((c = getc(fp)) != EOF && (strchr(WHITESPACE_CHARS, c) == NULL)) count++;

    
    while ((c = getc(fp)) != EOF) 
    {
        count++;    
        tempconfig = realloc (tempconfig, sizeof(char) * count);
        tempconfig[count - 1] = c;
    }
    count++;
    tempconfig = realloc (tempconfig, sizeof(char) * count);
    tempconfig[count - 1] = '\0';

/*
    tempconfig = (char *) malloc(count * sizeof(char) + 2);


    if ((fseek(fp, 0L, SEEK_SET)) != 0)
    {
        lcmaps_log(0, "%s: unable to reset the fp\n", logstr);
        fclose(fp);
        return 1;
    }

    if ((fgets(tempconfig, count+1, fp)) == NULL )
    {
        lcmaps_log(0, "%s: unable to read the configuration: fgets failed\n", logstr);
        fclose(fp);
        return 1;
    }
*/

    fclose(fp);

    lcmaps_log_debug(3,"%s: configuration: %s\n", logstr, tempconfig);

    if (tempconfig && (strlen (tempconfig) ))
    {
        parseConfigString (tempconfig);
    } 
    else
    {   
        lcmaps_log(0, "%s: no configuration extracted from file, no parsing will commence. Failure.\n", logstr);
        return 1;
    }

    // *jobrep_config = strdup(tempconfig);

    /* free mem */
    if (tempconfig) free(tempconfig);

    return 0;
}



/******************************************************************************
Function:   parseConfigString ()

Description:

Parameters:
    jobrep_config : variable to set the config string in.

Returns:
    0 on success.
    1 on failure
******************************************************************************/

int parseConfigString (char * configStr)
{
//char * jr_db_name
//char * jr_db_user     = NULL;
//char * jr_db_pwd

    char * value = NULL;
    char * searchString = NULL;
    int    i = 0;
    
    char * tmp;



    // Username
    value = malloc (sizeof(char) * 1024);
    for (i = 0; i < 1024; i++) value[i] = '\0';
    searchString = "username=\0";
    tmp = configStr;

    while ( tmp != '\0' )
    { 
        if (strncmp(tmp, searchString, strlen(searchString)) != 0) 
        {
            tmp = &tmp[1];
        }
        else
        {
            if (strlen(tmp) <= (strlen(searchString) + 1))
            {
                lcmaps_log(0, "Parsing error in %s\n", configStr);
                return 1;
            }
            else
            {
                tmp = &tmp[strlen(searchString)];

                int j = 0;
                while ((tmp[j] != '\n')
                    && (tmp[j] != ' ')
                    && (tmp[j] != '\t')
                    && (tmp[j] != ';')
                    && (tmp[j] != ':')
                    && (tmp[j] != '\0') )
                {
                    value[j] = tmp[j];
                    j++;
                }
                break;
            }
        }
    }
    
    jr_db_user = strdup(value);
    free(value);

    
    // Database
    value = malloc (sizeof(char) * 1024);
    for (i = 0; i < 1024; i++) value[i] = '\0';
    searchString = "password=\0";
    tmp = configStr;

    while ( tmp != '\0' )
    { 
        if (strncmp(tmp, searchString, strlen(searchString)) != 0) 
        {
            tmp = &tmp[1];
        }
        else
        {
            if (strlen(tmp) <= (strlen(searchString) + 1))
            {
                lcmaps_log(0, "Parsing error in %s\n", configStr);
                return 1;
            }
            else
            {
                tmp = &tmp[strlen(searchString)];

                int j = 0;
                while ((tmp[j] != '\n')
                    && (tmp[j] != ' ')
                    && (tmp[j] != '\t')
                    && (tmp[j] != ';')
                    && (tmp[j] != ':')
                    && (tmp[j] != '\0') )
                {
                    value[j] = tmp[j];
                    j++;
                }
                break;
            }
        }
    }

    jr_db_pwd = strdup(value);
    free(value);

   
 
    // Database
    value = malloc (sizeof(char) * 1024);
    for (i = 0; i < 1024; i++) value[i] = '\0';
    searchString = "database=\0";
    tmp = configStr;

    while ( tmp != '\0' )
    { 
        if (strncmp(tmp, searchString, strlen(searchString)) != 0) 
        {
            tmp = &tmp[1];
        }
        else
        {
            if (strlen(tmp) <= (strlen(searchString) + 1))
            {
                lcmaps_log(0, "Parsing error in %s\n", configStr);
                return 1;
            }
            else
            {
                tmp = &tmp[strlen(searchString)];

                int j = 0;
                while ((tmp[j] != '\n')
                    && (tmp[j] != ' ')
                    && (tmp[j] != '\t')
                    && (tmp[j] != ';')
                    && (tmp[j] != ':')
                    && (tmp[j] != '\0') )
                {
                    value[j] = tmp[j];
                    j++;
                }
                break;
            }
        }
    }

    
    jr_db_name = strdup(value);
    free(value);
    ///////////


    lcmaps_log_debug(1, "parse Config String : uname value -------------------------->>>>>>>>>>>>>>>>>>>>   | %s |\n", jr_db_user);
    lcmaps_log_debug(2, "parse Config String : pwd value ---------------------------->>>>>>>>>>>>>>>>>>>>   | %s |\n", jr_db_pwd);
    lcmaps_log_debug(1, "parse Config String : db name value ------------------------>>>>>>>>>>>>>>>>>>>>   | %s |\n", jr_db_name);


    return 0;
}


